@@ -28,8 +28,29 @@ HITTING0 = 'HITTING0'
 HITTING1 = 'HITTING1'
 CATAPULT0 = 'CATAPULT0'
 CATAPULT1 = 'CATAPULT1'
+WATERPISTOL = 'WATERPISTOL'
 BENDING_OVER = 'BENDING_OVER'
 FLY = 'FLY'
+RUN = 'RUN'
 SIT = 'SIT'
 HOP1 = 'HOP1'
 HOP2 = 'HOP2'
+WATER0 = 'WATER0'
+WATER1 = 'WATER1'
+WATER2 = 'WATER2'
+WATER3 = 'WATER3'
+WATER4 = 'WATER4'
+STINKBOMB = 'STINKBOMB'
+DESK_EMPTY = 'DESK_EMPTY'
+DESK_STINKBOMBS = 'DESK_STINKBOMBS'
+DESK_WATER_PISTOL = 'DESK_WATER_PISTOL'
+BIKE_ON_FLOOR = 'BIKE_ON_FLOOR'
+BIKE_UPRIGHT = 'BIKE_UPRIGHT'
+PLANT_GROWING = 'PLANT_GROWING'
+PLANT_GROWN = 'PLANT_GROWN'
+WATER_DROP = 'WATER_DROP'
+SHERRY_DROP = 'SHERRY_DROP'
+CONKER = 'CONKER'
+RIDING_BIKE0 = 'RIDING_BIKE0'
+RIDING_BIKE1 = 'RIDING_BIKE1'
+KISSING_ERIC = 'KISSING_ERIC'
