@@ -34,6 +34,8 @@ class Screen:
         self.score_box = gallery.get_score_box()
         self.logo = gallery.get_logo()
         self.font = Font(gallery.get_font())
+        self.inventory_coords = None
+        self.mouse_box_coords = None
 
     def scale_coords(self, coords):
         return (8 * self.scale * coords[0], 8 * self.scale * coords[1])
@@ -67,6 +69,8 @@ class Screen:
     def reinitialise(self):
         self.column = self.initial_column
         self.print_lesson('', '')
+        self.print_inventory()
+        self.print_mice()
 
     def add_score_box(self, coords, x_offset, attribute):
         self.score_box_coords = coords
@@ -164,7 +168,7 @@ class Screen:
     def print_hi_score(self, hiscore):
         self.print_number(hiscore, 17)
 
-    def print_inventory(self, item_images):
+    def print_inventory(self, item_images=()):
         if not self.inventory_coords:
             return
         inventory_box = pygame.Surface((self.scale * 48, self.scale * 8))
@@ -178,11 +182,11 @@ class Screen:
         self.screen.blit(inventory_box, self.scale_coords(self.inventory_coords))
         self.update()
 
-    def print_mice(self, count, mouse_image):
+    def print_mice(self, count=0, mouse_image=None):
         if not self.mouse_box_coords:
             return
         max_mice = 8
-        mouse_box = pygame.Surface((max_mice * mouse_image.get_width(), mouse_image.get_height()))
+        mouse_box = pygame.Surface(self.scale_coords((max_mice, 1)))
         ink, paper = self.get_ink_and_paper(self.background)
         mouse_box.fill(paper)
         for x in range(min((count, max_mice))):
@@ -190,9 +194,9 @@ class Screen:
         self.screen.blit(mouse_box, self.scale_coords(self.mouse_box_coords))
         self.update()
 
-    def contains(self, character):
-        """Returns whether a character is fully on-screen."""
-        return self.column <= character.x <= self.column + 29
+    def contains(self, character, full=True):
+        """Return whether a character is on-screen."""
+        return self.column <= character.x <= self.column + (29 if full else 31)
 
     def print_lines_bubble(self, x, y, message, ink_colour, paper_colour):
         bubble_x = 8 * (x // 8) - self.column
