@@ -18,7 +18,6 @@
 import pygame
 import random
 import items
-import debug
 
 class Skool:
     def __init__(self, game, screen, beeper, cast, timetable, gallery):
@@ -39,7 +38,9 @@ class Skool:
         self.locations = {}
         self.rooms = {}
         self.doors = {}
-        self.walls = []
+        self.windows = {}
+        self.walls = {}
+        self.barriers = {} # Doors, windows and walls
         self.staircases = {}
         self.floors = {}
         self.routes = {}
@@ -48,24 +49,33 @@ class Skool:
         self.shields = []
         self.shield_mode = 1
         self.safe_combination = None
+        self.bike_combination = None
+        self.storeroom_combination = None
         self.draw_index = 0
         self.signals = Signals()
         self.messages = {}
         self.expelled = False
+        self.cups = {}
+        self.plant_pots = {}
+        self.bike = None
+        self.clear_score = True
 
     def add_location(self, location_id, coords):
         self.locations[location_id] = coords
 
-    def add_room(self, room_id, name, y, min_x, max_x):
-        self.rooms[room_id] = Room(room_id, name, y, min_x, max_x)
+    def add_room(self, room_id, name, y, min_x, max_x, get_along):
+        self.rooms[room_id] = Room(room_id, name, y, min_x, max_x, get_along)
 
     def add_chair(self, room_id, x):
         self.rooms[room_id].add_chair(x)
 
-    def add_door(self, door_id, x, bottom_y, top_y, shut, auto_shuts):
-        self.doors[door_id] = Door(door_id, x, bottom_y, top_y, shut, auto_shuts)
+    def add_desk(self, room_id, x):
+        self.rooms[room_id].add_desk(x)
 
-    def add_door_images(self, door_id, shut_top_left, size, coords):
+    def add_door(self, door_id, x, bottom_y, top_y, initially_shut, auto_shuts, shut_top_left, size, coords):
+        door = Door(door_id, x, bottom_y, top_y, initially_shut, auto_shuts)
+        self.doors[door_id] = door
+        self.barriers[door_id] = door
         size = self.scale_coords(size)
         shut_top_left = self.scale_coords(shut_top_left)
         open_top_left = (shut_top_left[0] + size[0], shut_top_left[1])
@@ -79,7 +89,26 @@ class Skool:
             open_image_ink = self.mutables_ink.subsurface(open_top_left, size)
             open_image_paper = self.mutables_paper.subsurface(open_top_left, size)
             open_images = (open_image_ink, open_image_paper)
-        self.doors[door_id].set_images(open_images, shut_images, coords)
+        door.set_images(open_images, shut_images, coords)
+
+    def add_window(self, window_id, x, bottom_y, top_y, initially_shut, opener_coords, shut_top_left, size, coords):
+        window = Window(window_id, x, bottom_y, top_y, initially_shut, opener_coords)
+        self.windows[window_id] = window
+        self.barriers[window_id] = window
+        size = self.scale_coords(size)
+        shut_top_left = self.scale_coords(shut_top_left)
+        open_top_left = (shut_top_left[0] + size[0], shut_top_left[1])
+        if self.mode == 0:
+            shut_images = (self.mutables.subsurface(shut_top_left, size),)
+            open_images = (self.mutables.subsurface(open_top_left, size),)
+        else:
+            shut_image_ink = self.mutables_ink.subsurface(shut_top_left, size)
+            shut_image_paper = self.mutables_paper.subsurface(shut_top_left, size)
+            shut_images = (shut_image_ink, shut_image_paper)
+            open_image_ink = self.mutables_ink.subsurface(open_top_left, size)
+            open_image_paper = self.mutables_paper.subsurface(open_top_left, size)
+            open_images = (open_image_ink, open_image_paper)
+        window.set_images(open_images, shut_images, coords)
 
     def add_inventory_item(self, item_id, top_left, size):
         top_left = self.scale_coords(top_left)
@@ -97,8 +126,10 @@ class Skool:
     def get_mouse_image(self):
         return self.inventory_images.get(items.MOUSE)
 
-    def add_wall(self, x, bottom_y, top_y):
-        self.walls.append(Wall(x, bottom_y, top_y))
+    def add_wall(self, wall_id, x, bottom_y, top_y):
+        wall = Wall(wall_id, x, bottom_y, top_y)
+        self.walls[wall_id] = wall
+        self.barriers[wall_id] = wall
 
     def add_staircase(self, staircase_id, bottom, top, force, alias):
         staircase = Staircase(bottom, top, force)
@@ -137,6 +168,51 @@ class Skool:
     def add_safe(self, x, y, score, image_index):
         self.safe = self.get_flashable(Safe, x, y, score, image_index)
 
+    def add_cup(self, cup_id, empty_top_left, size, coords):
+        cup = Cup(cup_id, coords)
+        self.cups[cup_id] = cup
+        size = self.scale_coords(size)
+        empty_top_left = self.scale_coords(empty_top_left)
+        water_top_left = (empty_top_left[0] + size[0], empty_top_left[1])
+        sherry_top_left = (water_top_left[0] + size[0], empty_top_left[1])
+        if self.mode == 0:
+            empty_images = (self.mutables.subsurface(empty_top_left, size),)
+            water_images = (self.mutables.subsurface(water_top_left, size),)
+            sherry_images = (self.mutables.subsurface(sherry_top_left, size),)
+        else:
+            empty_image_ink = self.mutables_ink.subsurface(empty_top_left, size)
+            empty_image_paper = self.mutables_paper.subsurface(empty_top_left, size)
+            empty_images = (empty_image_ink, empty_image_paper)
+            water_image_ink = self.mutables_ink.subsurface(water_top_left, size)
+            water_image_paper = self.mutables_paper.subsurface(water_top_left, size)
+            water_images = (water_image_ink, water_image_paper)
+            sherry_image_ink = self.mutables_ink.subsurface(sherry_top_left, size)
+            sherry_image_paper = self.mutables_paper.subsurface(sherry_top_left, size)
+            sherry_images = (sherry_image_ink, sherry_image_paper)
+        cup.set_images(empty_images, water_images, sherry_images)
+
+    def add_plant(self, plant_id, sprite_group_id, x, y, tap_id):
+        plant = self.cast.add_plant(plant_id, sprite_group_id, x, y, tap_id)
+        self.plant_pots[plant_id] = PlantPot(plant_id, plant, x, y - 1)
+
+    def add_bike(self, bike_id, sprite_group_id, initial_as, location, tap_id, top_left, size, coords):
+        self.cast.add_bike(bike_id, sprite_group_id, initial_as, location, tap_id)
+        self.bike = Bike(*coords)
+        size = self.scale_coords(size)
+        unchained_top_left = self.scale_coords(top_left)
+        chained_top_left = (unchained_top_left[0] + size[0], unchained_top_left[1])
+        if self.mode == 0:
+            unchained_images = (self.mutables.subsurface(unchained_top_left, size),)
+            chained_images = (self.mutables.subsurface(chained_top_left, size),)
+        else:
+            unchained_image_ink = self.mutables_ink.subsurface(unchained_top_left, size)
+            unchained_image_paper = self.mutables_paper.subsurface(unchained_top_left, size)
+            unchained_images = (unchained_image_ink, unchained_image_paper)
+            chained_image_ink = self.mutables_ink.subsurface(chained_top_left, size)
+            chained_image_paper = self.mutables_paper.subsurface(chained_top_left, size)
+            chained_images = (chained_image_ink, chained_image_paper)
+        self.bike.set_images(unchained_images, chained_images)
+
     def add_message(self, message_id, message):
         self.messages[message_id] = message
 
@@ -164,16 +240,28 @@ class Skool:
                         self.add_to_score(200)
                         self.beeper.play_all_shields_tune()
                     elif self.shield_mode == 3 and self.all_shields_unflashed():
-                        self.shield_mode = 1
-                        self.safe.unflash()
-                        self.add_to_score(500)
-                        message_lines = self.messages['NEXT_YEAR'].split('^')
-                        self.screen.print_lesson(*message_lines)
-                        self.initialise_secrets()
-                        self.beeper.play_tune()
-                        self.timetable.up_a_year()
+                        self.up_a_year('SkoolDaze')
                 return True
 
+    def up_a_year(self, game_id):
+        if game_id == 'SkoolDaze':
+            self.shield_mode = 1
+            self.safe.unflash()
+            self.add_to_score(500)
+            self.print_up_a_year_message()
+            self.initialise_safe_combo()
+            self.beeper.play_tune()
+            self.timetable.up_a_year()
+        elif game_id == 'BackToSkool':
+            self.add_to_score(self.safe.score // 10)
+            self.print_up_a_year_message()
+            self.beeper.play_up_a_year_tune()
+            self.end_game(False)
+
+    def print_up_a_year_message(self):
+        message_lines = self.messages['NEXT_YEAR'].split('^')
+        self.screen.print_lesson(*message_lines)
+
     def get_blackboards(self):
         return [room.blackboard for room in self.rooms.values() if room.blackboard]
 
@@ -182,14 +270,30 @@ class Skool:
             if board.shows(self.safe_combination):
                 return True
 
-    def check_safe(self, x, y):
-        if self.shield_mode == 2 and (self.safe.x, self.safe.y) == (x, y) and self.got_combination():
-            self.shield_mode += 1
-            self.safe.flash()
-            self.add_to_score(100)
-            self.beeper.play_all_shields_tune()
+    def check_safe(self, x, y, got_key):
+        if (self.safe.x, self.safe.y) == (x, y):
+            if got_key:
+                # Must be the Back to Skool safe
+                self.up_a_year('BackToSkool')
+            elif self.shield_mode == 2 and self.got_combination():
+                self.shield_mode += 1
+                self.safe.flash()
+                self.add_to_score(100)
+                self.beeper.play_all_shields_tune()
 
-    def can_reveal_secret(self):
+    def check_drinks_cabinet(self, x, y):
+        door = self.doors.get('DrinksCabinet')
+        if door:
+            return (x, y) == door.top_left and not door.shut
+        return False
+
+    def got_bike_combination(self, blackboard):
+        return blackboard.shows(self.bike_combination, False)
+
+    def got_storeroom_combination(self, blackboard):
+        return blackboard.shows(self.storeroom_combination, False)
+
+    def can_reveal_safe_secret(self):
         return self.shield_mode == 2
 
     def get_room(self, room_id):
@@ -228,35 +332,98 @@ class Skool:
         staircase_id = routes.get(dest_floor.floor_id, routes['*'])
         return self.staircases[staircase_id]
 
-    def floor(self, location):
+    def floor(self, thing):
         for floor in self.floors.values():
-            if floor.supports(location):
+            if floor.supports(thing):
                 return floor
 
+    def floor_below(self, character):
+        floor = None
+        for f in self.floors.values():
+            if f.below(character):
+                if floor is None or f.y < floor.y:
+                    floor = f
+        return floor
+
     def on_floor(self, character):
         return any(floor.supports(character) for floor in self.floors.values())
 
+    def floor_at(self, x, y):
+        for floor in self.floors.values():
+            if floor.contains_location(x, y):
+                return floor
+
     def in_no_go_zone(self, character):
         return any(zone.contains(character) for zone in self.no_go_zones)
 
     def barrier(self, character, distance=0):
-        for barrier in self.walls + self.doors.values():
+        for barrier in self.barriers.values():
             if barrier.impedes(character, distance):
                 return barrier
 
-    def chair(self, character):
+    def chair(self, character, direction=-1):
         room = self.room(character)
         if room:
-            return room.chair(character)
+            return room.chair(character, direction)
+
+    def desk(self, character):
+        room = self.room(character)
+        if room:
+            return room.desk(character)
+
+    def cup(self, x, y):
+        for cup in self.cups.values():
+            if (cup.x, cup.y) == (x, y):
+                return cup
+
+    def plant_pot(self, x, y):
+        for plant_pot in self.plant_pots.values():
+            if (plant_pot.x, plant_pot.y) == (x, y):
+                return plant_pot
+
+    def window(self, character):
+        """Returns the window (if any) in front of a character."""
+        for window in self.windows.values():
+            if window.impedes(character, force_shut=True):
+                return window
+
+    def nearby_window(self, character):
+        for window in self.windows.values():
+            x, y = window.opener_coords
+            if character.y == y and abs(character.x - x) <= 16:
+                return window
 
     def scale_coords(self, coords):
         return (8 * self.scale * coords[0], 8 * self.scale * coords[1])
 
-    def is_door_shut(self, door_id):
-        return self.doors[door_id].is_shut()
+    def is_door_shut(self, barrier_id):
+        return self.barriers[barrier_id].is_shut()
 
-    def move_door(self, door_id, shut):
-        images, coords = self.doors[door_id].move(shut)
+    def move_door(self, barrier_id, shut):
+        images, coords = self.barriers[barrier_id].move(shut)
+        self.draw_mutable(images, coords)
+
+    def auto_shut_doors(self):
+        for barrier_id, door in self.doors.items():
+            if door.auto_shut() and not self.cast.somebody_near_door(door):
+                self.move_door(barrier_id, True)
+
+    def fill_cup(self, cup_id, contents):
+        images, coords = self.cups[cup_id].fill(contents)
+        self.draw_mutable(images, coords)
+
+    def unchain_bike(self):
+        if self.bike:
+            images, coords = self.bike.unchain()
+            self.draw_mutable(images, coords)
+            self.cast.unchain_bike()
+
+    def chain_bike(self):
+        if self.bike:
+            images, coords = self.bike.chain()
+            self.draw_mutable(images, coords)
+
+    def draw_mutable(self, images, coords):
         scaled_coords = self.scale_coords(coords)
         if self.mode == 0:
             self.skool.blit(images[0], scaled_coords)
@@ -264,11 +431,6 @@ class Skool:
             self.ink.blit(images[0], scaled_coords)
             self.paper.blit(images[1], scaled_coords)
 
-    def auto_shut_doors(self):
-        for door in self.doors.values():
-            if door.auto_shut() and not self.cast.somebody_near_door(door):
-                self.move_door(door.door_id, True)
-
     def resolve_location_id(self, location_id):
         location_marker = 'Location:'
         if location_id.startswith(location_marker):
@@ -307,55 +469,93 @@ class Skool:
         return self.assembly_message_generator.generate_message()
 
     def line_of_sight_between(self, a, b):
-        """Returns whether there is a clear line of sight between a and b (i.e.
-        there are no walls between them."""
-        return not any(wall.separates(a, b) for wall in self.walls)
+        """Return whether there is a clear line of sight between two points (i.e. there are no walls between them)."""
+        return not any(wall.separates(a, b) for wall in self.walls.values())
 
     def wipe_board(self, blackboard, column):
-        if blackboard is None:
-            debug.log('Cannot wipe non-existent blackboard')
-            return
-        if column == 0:
-            blackboard.clear()
-            return
-        wiped_bit = pygame.Surface((8 * self.scale, 8 * self.scale))
-        blackboard.wipe(wiped_bit, 8 * column * self.scale)
+        if blackboard:
+            if column == 0:
+                blackboard.clear()
+                return
+            wiped_bit = pygame.Surface((8 * self.scale, 8 * self.scale))
+            blackboard.wipe(wiped_bit, 8 * column * self.scale)
 
     def write_on_board(self, character, blackboard, message, index=1):
-        if blackboard is None:
-            debug.log('Cannot write on non-existent blackboard')
-            return
-        blackboard.write(message[index - 1])
-        blackboard.set_writer(character)
+        if blackboard:
+            blackboard.write(message[index - 1])
+            blackboard.set_writer(character)
         return index == len(message)
 
-    def initialise_secrets(self):
-        letters = self.cast.initialise_secrets()
+    def initialise_safe_combo(self):
+        letters = self.cast.initialise_safe_combo()
         if letters:
             self.safe_combination = letters[0]
             indexes = [n for n in range(1, len(letters))]
             while indexes:
                 self.safe_combination += letters[indexes.pop(random.randrange(len(indexes)))]
 
+    def initialise_bike_combo(self):
+        self.bike_combination = self.cast.initialise_bike_combo()
+
+    def initialise_storeroom_combo(self):
+        self.storeroom_combination = self.cast.initialise_storeroom_combo()
+
     def initialise_cast(self):
         self.cast.initialise(self, self.screen, self.beeper)
-        self.initialise_secrets()
+        self.initialise_safe_combo()
+        self.initialise_bike_combo()
+        self.initialise_storeroom_combo()
 
     def reinitialise(self):
         self.expelled = False
         self.screen.reinitialise()
         self.signals.clear()
         self.cast.reinitialise()
-        self.initialise_secrets()
+        self.initialise_safe_combo()
+        self.initialise_bike_combo()
+        self.initialise_storeroom_combo()
         self.timetable.reinitialise()
-        self.scoreboard.reinitialise()
+        if self.clear_score:
+            self.scoreboard.reinitialise()
         for room in self.rooms.values():
             room.wipe_blackboard()
         for shield in self.shields:
             shield.unflash()
+        self.fill_desks()
+        for cup_id in self.cups:
+            self.fill_cup(cup_id, None)
+        for door in self.doors.values():
+            self.move_door(door.barrier_id, door.initially_shut)
+        for window in self.windows.values():
+            self.move_door(window.barrier_id, door.initially_shut)
+        self.chain_bike()
+
+    def get_desks(self):
+        desks = []
+        for room in self.rooms.values():
+            desks.extend(room.desks)
+        return desks
+
+    def hide_in_desk(self, item):
+        desks = self.get_desks()
+        while desks:
+            desk = desks.pop(random.randrange(len(desks)))
+            if not desk.contents:
+                desk.insert(item)
+                break
+
+    def fill_desks(self):
+        desks = self.get_desks()
+        if not desks:
+            return
+        for desk in desks:
+            desk.empty()
+        desks.pop(random.randrange(len(desks))).insert(items.WATER_PISTOL)
+        if desks:
+            random.choice(desks).insert(items.STINKBOMBS3)
 
     def get_eric(self):
-        return self.cast.get_eric()
+        return self.cast.eric
 
     def get_character(self, character_id):
         return self.cast.get(character_id)
@@ -369,6 +569,9 @@ class Skool:
     def is_time_to_start_lesson(self):
         return self.timetable.is_time_to_start_lesson()
 
+    def is_time_remaining(self, time):
+        return self.timetable.is_time_remaining(time)
+
     def is_teaching_eric(self, character):
         return self.timetable.is_teaching_eric(character)
 
@@ -386,14 +589,16 @@ class Skool:
         return self.home_room
 
     def should_get_along(self, eric):
-        """Returns whether Eric is somewhere other than he should be."""
+        """Return whether Eric is somewhere other than he should be."""
         if self.home_room:
             return not self.home_room.contains(eric)
-        destination = self.get_room(self.timetable.get_room_id())
-        dinner_hall = self.rooms.get('DinnerHall', None)
         if self.timetable.is_time(200):
             # Give Eric time to leave a classroom
-            return self.room(eric) not in (None, dinner_hall, destination)
+            room = self.room(eric)
+            if room:
+                destination = self.get_room(self.timetable.get_room_id())
+                return room != destination and room.get_along
+        return False
 
     def get_lesson_desc(self):
         teacher_id = self.timetable.get_teacher_id()
@@ -448,6 +653,9 @@ class Skool:
     def start_clock(self, ticks):
         self.timetable.resume(ticks)
 
+    def rewind_clock(self, ticks):
+        self.timetable.rewind(ticks)
+
     def get_too_many_lines_message(self):
         return self.cast.expand_names(self.messages['TOO_MANY_LINES'])
 
@@ -463,7 +671,8 @@ class Skool:
     def got_signal(self, signal):
         return self.signals.is_raised(signal)
 
-    def end_game(self):
+    def end_game(self, clear_score=True):
+        self.clear_score = clear_score
         self.game.end()
 
 class Location:
@@ -471,6 +680,9 @@ class Location:
         self.x = coords[0]
         self.y = coords[1]
 
+    def coords(self):
+        return self.x, self.y
+
     def __str__(self):
         return '(%s, %s)' % (self.x, self.y)
 
@@ -486,7 +698,7 @@ class Staircase:
     def contains(self, character, distance=0):
         """
         Returns whether the character is one of the following:
-            i) on this staircase
+            i) on a step of this staircase
            ii) at the bottom of this staircase facing the top
           iii) at the top of this staircase facing the bottom
         """
@@ -502,18 +714,19 @@ class Staircase:
     def contains_location(self, x, y):
         """
         Returns whether the given location (x, y) is at the bottom, at the top,
-        or on this staircase.
+        or on a step of this staircase.
         """
-        nx = self.direction * x
-        if self.direction * self.bottom.x <= nx <= self.direction * self.top.x:
-            return y == self.bottom.y - abs(x - self.bottom.x)
+        if self.direction * self.bottom.x <= self.direction * x <= self.direction * self.top.x:
+            if (x, y) in (self.bottom.coords(), self.top.coords()):
+                return True
+            return self.top.y < y < self.bottom.y
         return False
 
     def supports(self, character):
         """Returns whether the character is on a step of this staircase."""
         if not self.contains_location(character.x, character.y):
             return False
-        return character.x not in (self.bottom.x, self.top.x)
+        return (character.x, character.y) not in (self.bottom.coords(), self.top.coords())
 
 class Floor:
     def __init__(self, floor_id, left_x, right_x, y):
@@ -525,28 +738,35 @@ class Floor:
     def supports(self, character):
         return self.y == character.y and self.left_x <= character.x <= self.right_x
 
+    def below(self, character):
+        return self.y >= character.y and self.left_x <= character.x <= self.right_x
+
+    def contains_location(self, x, y):
+        return self.y == y and self.left_x <= x <= self.right_x
+
 class Barrier:
-    def __init__(self, x, bottom_y, top_y):
+    def __init__(self, barrier_id, x, bottom_y, top_y):
+        self.barrier_id = barrier_id
         self.x = x
         self.bottom_y = bottom_y
         self.top_y = top_y
 
-    def impedes(self, character, distance=0):
-        if not self.is_shut():
+    def impedes(self, character, distance=0, force_shut=False):
+        if not (force_shut or self.is_shut()):
             return False
-        if self.bottom_y >= character.y >= self.top_y:
+        if character.impeded(self.bottom_y, self.top_y):
             if character.x < self.x and character.direction > 0:
                 return character.x + distance >= self.x - 2
             elif character.x >= self.x and character.direction < 0:
                 return character.x - distance <= self.x
         return False
 
-    def is_openable(self):
-        return False
-
     def is_shut(self):
         return True
 
+    def is_door(self):
+        return False
+
 class Wall(Barrier):
     def separates(self, a, b):
         """Returns whether this wall blocks the view from a to b."""
@@ -556,10 +776,9 @@ class Wall(Barrier):
             return self.top_y <= max(a.y, b.y) and self.bottom_y >= min(a.y, b.y)
 
 class Door(Barrier):
-    def __init__(self, door_id, x, bottom_y, top_y, shut, auto_shuts=True):
-        Barrier.__init__(self, x, bottom_y, top_y)
-        self.door_id = door_id
-        self.shut = shut
+    def __init__(self, door_id, x, bottom_y, top_y, shut, auto_shuts):
+        Barrier.__init__(self, door_id, x, bottom_y, top_y)
+        self.shut = self.initially_shut = shut
         self.auto_shuts = auto_shuts
         self.open_images = None
         self.shut_images = None
@@ -567,15 +786,9 @@ class Door(Barrier):
         self.auto_shut_timer = 0
         self.auto_shut_delay = 40
 
-    def impedes(self, character, distance=0):
-        return self.shut and Barrier.impedes(self, character, distance)
-
     def is_shut(self):
         return self.shut
 
-    def is_openable(self):
-        return True
-
     def move(self, shut):
         self.shut = shut
         if not shut:
@@ -593,14 +806,24 @@ class Door(Barrier):
             self.auto_shut_timer -= 1
         return self.auto_shut_timer < 0
 
+    def is_door(self):
+        return True
+
+class Window(Door):
+    def __init__(self, window_id, x, bottom_y, top_y, shut, opener_coords):
+        Door.__init__(self, window_id, x, bottom_y, top_y, shut, False)
+        self.opener_coords = opener_coords
+
 class Room:
-    def __init__(self, room_id, name, y, min_x, max_x):
+    def __init__(self, room_id, name, y, min_x, max_x, get_along):
         self.room_id = room_id
         self.name = name
         self.y = y
         self.min_x = min_x
         self.max_x = max_x
+        self.get_along = get_along
         self.chairs = []
+        self.desks = []
         self.blackboard = None
 
     def get_blackboard(self):
@@ -632,6 +855,9 @@ class Room:
     def add_chair(self, x):
         self.chairs.append(Chair(self, x))
 
+    def add_desk(self, x):
+        self.desks.append(Desk(self, x))
+
     def beside_blackboard(self, character):
         if self.blackboard:
             return self.y == character.y and self.blackboard.beside(character)
@@ -639,11 +865,16 @@ class Room:
     def contains(self, character):
         return self.y - 3 <= character.y <= self.y and self.min_x <= character.x <= self.max_x
 
-    def chair(self, character):
+    def chair(self, character, direction=-1):
         for chair in self.chairs:
-            if character.direction < 0 and (character.x, character.y) == (chair.x, chair.y):
+            if (direction is None or character.direction == direction) and (character.x, character.y) == (chair.x, chair.y):
                 return chair
-        return None
+
+    def desk(self, character):
+        if character.is_sitting_on_chair():
+            for desk in self.desks:
+                if (character.x, character.y) == (desk.x, desk.y):
+                    return desk
 
     def get_next_chair(self, character, move_along):
         """Returns the chair nearest to the character, or the chair he should
@@ -675,6 +906,19 @@ class Chair:
     def vacate(self):
         self.occupant = None
 
+class Desk:
+    def __init__(self, room, x):
+        self.x = x
+        self.y = room.y
+        self.room = room
+        self.empty()
+
+    def insert(self, item):
+        self.contents = item
+
+    def empty(self):
+        self.contents = None
+
 class Blackboard:
     def __init__(self, x, y, font, scale):
         self.x = x
@@ -721,8 +965,10 @@ class Blackboard:
         arm_x = character.x + character.direction + 1
         return self.x + 1 <= arm_x <= self.x + 6
 
-    def shows(self, text):
-        return self.lines[0].lower().startswith(text.lower())
+    def shows(self, text, in_order=True):
+        if in_order:
+            return self.lines[0].lower().startswith(text.lower())
+        return set(text.lower()) == set(self.lines[0][:len(text)].lower())
 
 class NoGoZone:
     def __init__(self, zone_id, min_x, max_x, bottom_y, top_y):
@@ -770,6 +1016,61 @@ class Shield(Flashable):
 class Safe(Flashable):
     pass
 
+class Cup:
+    def __init__(self, cup_id, coords):
+        self.cup_id = cup_id
+        self.x, self.y = coords
+        self.contents = None
+        self.frogs = []
+
+    def set_images(self, empty_images, water_images, sherry_images):
+        self.empty_images = empty_images
+        self.water_images = water_images
+        self.sherry_images = sherry_images
+
+    def is_empty(self):
+        return not (self.contents or self.frogs)
+
+    def fill(self, contents):
+        self.contents = contents
+        if contents == 'WATER':
+            images = self.water_images
+        elif contents == 'SHERRY':
+            images = self.sherry_images
+        else:
+            images = self.empty_images
+        return images, (self.x, self.y)
+
+    def insert_frog(self, frog):
+        self.frogs.append(frog)
+        frog.insert_into_cup(self)
+
+    def remove_frog(self, frog):
+        if frog in self.frogs:
+            self.frogs.remove(frog)
+
+class PlantPot:
+    def __init__(self, plant_id, plant, x, y):
+        self.plant_id = plant_id
+        self.plant = plant
+        self.x = x
+        self.y = y
+
+class Bike:
+    def __init__(self, x, y):
+        self.x = x
+        self.y = y
+
+    def set_images(self, unchained_images, chained_images):
+        self.unchained_images = unchained_images
+        self.chained_images = chained_images
+
+    def unchain(self):
+        return self.unchained_images, (self.x, self.y)
+
+    def chain(self):
+        return self.chained_images, (self.x, self.y)
+
 class Timetable:
     def __init__(self, lesson_length, lesson_start_time):
         self.index = -1
@@ -824,6 +1125,9 @@ class Timetable:
     def is_time(self, time):
         return self.counter + time < self.lesson_length
 
+    def is_time_remaining(self, time):
+        return self.counter < time
+
     def is_time_to_start_lesson(self):
         return self.is_time(self.lesson_start_time)
 
@@ -840,6 +1144,9 @@ class Timetable:
         self.counter = ticks
         self.ticking = True
 
+    def rewind(self, ticks):
+        self.counter += ticks
+
 class AssemblyMessageGenerator:
     def __init__(self):
         self.text = None
@@ -875,7 +1182,7 @@ class Scoreboard:
         self.screen.print_score(self.score)
 
     def add_lines(self, addend):
-        self.lines += addend
+        self.lines = max(self.lines + addend, 0)
         self.screen.print_lines(self.lines)
 
     def reinitialise(self):
